@@ -25,48 +25,47 @@
 # OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 import datetime
-import random
-import signal
 import socket
-import sys
 from statistics import stdev
+from typing import Optional, List, Any
 
 import httpx
-import dns.flags
+import dns.edns
 import dns.message
 import dns.query
 import dns.rcode
 import dns.rdataclass
-import string
 
-shutdown = False
+from dnsdiag.shared import random_string, die, err, unsupported_feature
 
 # Transport protocols
-PROTO_UDP = 0
-PROTO_TCP = 1
-PROTO_TLS = 2
-PROTO_HTTPS = 3
-PROTO_QUIC = 4
-PROTO_HTTP3 = 5
+PROTO_UDP: int = 0
+PROTO_TCP: int = 1
+PROTO_TLS: int = 2
+PROTO_HTTPS: int = 3
+PROTO_QUIC: int = 4
+PROTO_HTTP3: int = 5
 
-_TTL = None
+_TTL: Optional[int] = None
 
 
 class PingResponse:
-    def __init__(self):
-        self.r_avg = 0
-        self.r_min = 0
-        self.r_max = 0
-        self.r_stddev = 0
-        self.r_lost_percent = 0
-        self.flags = 0
-        self.ttl = None
-        self.answer = None
-        self.rcode = 0
-        self.rcode_text = ''
+    def __init__(self) -> None:
+        self.r_avg: float = 0.0
+        self.r_min: float = 0.0
+        self.r_max: float = 0.0
+        self.r_stddev: float = 0.0
+        self.r_lost_percent: float = 0.0
+        self.flags: int = 0
+        self.ednsflags: int = 0
+        self.ttl: Optional[int] = None
+        self.answer: Optional[Any] = None
+        self.rcode: int = 0
+        self.rcode_text: str = ''
+        self.response: Optional[Any] = None
 
 
-def proto_to_text(proto):
+def proto_to_text(proto: int) -> str:
     _proto_name = {
         PROTO_UDP: 'UDP',
         PROTO_TCP: 'TCP',
@@ -78,7 +77,7 @@ def proto_to_text(proto):
     return _proto_name[proto]
 
 
-def getDefaultPort(proto):
+def get_default_port(proto: int) -> int:
     _proto_port = {
         PROTO_UDP: 53,
         PROTO_TCP: 53,
@@ -91,19 +90,23 @@ def getDefaultPort(proto):
 
 
 class CustomSocket(socket.socket):
-    def __init__(self, *args, **kwargs):
+    def __init__(self, *args: Any, **kwargs: Any) -> None:
         super(CustomSocket, self).__init__(*args, **kwargs)
         if _TTL:
-            self.setsockopt(socket.SOL_IP, socket.IP_TTL, _TTL)
+            # Set TTL/hop limit based on address family
+            if self.family == socket.AF_INET:
+                self.setsockopt(socket.SOL_IP, socket.IP_TTL, _TTL)
+            elif self.family == socket.AF_INET6:
+                self.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_UNICAST_HOPS, _TTL)
 
 
-def ping(qname, server, dst_port, rdtype, timeout, count, proto, src_ip, use_edns=False, force_miss=False,
-         want_dnssec=False, socket_ttl=None):
+def ping(qname: str, server: str, dst_port: int, rdtype: str, timeout: float, count: int, proto: int,
+         src_ip: Optional[str], use_edns: bool = False, force_miss: bool = False,
+         want_dnssec: bool = False, want_nsid: bool = False, socket_ttl: Optional[int] = None) -> PingResponse:
     retval = PingResponse()
     retval.rcode_text = "No Response"
 
-    response_times = []
-    i = 0
+    response_times: List[float] = []
 
     if socket_ttl:
         global _TTL
@@ -112,47 +115,51 @@ def ping(qname, server, dst_port, rdtype, timeout, count, proto, src_ip, use_edn
 
     for i in range(count):
 
-        if shutdown:  # user pressed CTRL+C
-            raise SystemExit
-
         if force_miss:
             fqdn = "_dnsdiag_%s_.%s" % (random_string(), qname)
         else:
             fqdn = qname
 
         if use_edns:
-            query = dns.message.make_query(fqdn, rdtype, dns.rdataclass.IN, use_edns, want_dnssec, payload=1232)
+            edns_options: List[dns.edns.Option] = []
+            if want_nsid:
+                edns_options.append(dns.edns.GenericOption(dns.edns.NSID, b''))
+            query = dns.message.make_query(fqdn, rdtype, dns.rdataclass.IN, use_edns, want_dnssec, payload=1232,
+                                           options=edns_options if edns_options else None)
         else:
             query = dns.message.make_query(fqdn, rdtype, dns.rdataclass.IN, use_edns=False, want_dnssec=False)
 
         try:
-            if proto is PROTO_UDP:
+            if proto == PROTO_UDP:
                 response = dns.query.udp(query, server, timeout=timeout, port=dst_port, source=src_ip,
                                          ignore_unexpected=True)
-            elif proto is PROTO_TCP:
+            elif proto == PROTO_TCP:
                 response = dns.query.tcp(query, server, timeout=timeout, port=dst_port, source=src_ip)
-            elif proto is PROTO_TLS:
+            elif proto == PROTO_TLS:
                 if hasattr(dns.query, 'tls'):
                     response = dns.query.tls(query, server, timeout, dst_port, src_ip)
                 else:
                     unsupported_feature()
-            elif proto is PROTO_HTTPS:
+            elif proto == PROTO_HTTPS:
                 if hasattr(dns.query, 'https'):
-                    response = dns.query.https(query, server, timeout, dst_port, src_ip)
+                    response = dns.query.https(query, server, timeout, dst_port, src_ip,
+                                              http_version=dns.query.HTTPVersion.HTTP_2)
                 else:
                     unsupported_feature()
-            elif proto is PROTO_QUIC:
+            elif proto == PROTO_QUIC:
                 if hasattr(dns.query, 'quic'):
                     response = dns.query.quic(query, server, timeout, dst_port, src_ip)
                 else:
                     unsupported_feature()
-            elif proto is PROTO_HTTP3:
+            elif proto == PROTO_HTTP3:
                 if hasattr(dns.query, '_http3'):
                     url = f"https://{server}:{dst_port}/dns-query"
                     response = dns.query._http3(query, server, url, timeout, dst_port, src_ip)
                 else:
                     unsupported_feature()
 
+        except dns.query.NoDOH:
+            raise
         except (httpx.ConnectTimeout, httpx.ReadTimeout,
                 httpx.ConnectError):
             raise ConnectionError('Connection failed')
@@ -162,23 +169,31 @@ def ping(qname, server, dst_port, rdtype, timeout, count, proto, src_ip, use_edn
         except dns.exception.Timeout:
             break
         except OSError as e:
-            if socket_ttl:  # this is an acceptable error while doing traceroute
+            # Check for fatal network errors
+            if e.errno == 65:  # EHOSTUNREACH
+                die("ERROR: No route to host")
+            elif e.errno == 51:  # ENETUNREACH
+                die("ERROR: Network unreachable")
+            elif socket_ttl:  # other acceptable errors while doing traceroute
                 break
-            print("error: %s" % e.strerror, file=sys.stderr, flush=True)
+            err(f"ERROR: {e.strerror}")
             raise OSError(e)
         except Exception as e:
-            print("error: %s" % e, file=sys.stderr, flush=True)
+            err(f"ERROR: {e}")
             break
         else:
             # convert time to milliseconds, considering that
             # time property is retruned differently by query.https
-            if type(response.time) is datetime.timedelta:
+            # dns library returns float for most protocols but timedelta for HTTPS
+            if isinstance(response.time, datetime.timedelta):
                 elapsed = response.time.total_seconds() * 1000
             else:
                 elapsed = response.time * 1000
             response_times.append(elapsed)
             if response:
+                retval.response = response
                 retval.flags = response.flags
+                retval.ednsflags = response.ednsflags
                 retval.answer = response.answer
                 retval.rcode = response.rcode()
                 retval.rcode_text = dns.rcode.to_text(response.rcode())
@@ -187,8 +202,7 @@ def ping(qname, server, dst_port, rdtype, timeout, count, proto, src_ip, use_edn
 
     r_sent = i + 1
     r_received = len(response_times)
-    retval.r_lost_count = r_sent - r_received
-    retval.r_lost_percent = (100 * retval.r_lost_count) / r_sent
+    retval.r_lost_percent = (100 * (r_sent - r_received)) / r_sent
     if response_times:
         retval.r_min = min(response_times)
         retval.r_max = max(response_times)
@@ -196,39 +210,17 @@ def ping(qname, server, dst_port, rdtype, timeout, count, proto, src_ip, use_edn
         if len(response_times) > 1:
             retval.r_stddev = stdev(response_times)
         else:
-            retval.r_stddev = 0
+            retval.r_stddev = 0.0
     else:
-        retval.r_min = 0
-        retval.r_max = 0
-        retval.r_avg = 0
-        retval.r_stddev = 0
+        retval.r_min = 0.0
+        retval.r_max = 0.0
+        retval.r_avg = 0.0
+        retval.r_stddev = 0.0
 
     return retval
 
 
-def random_string(min_length=5, max_length=10):
-    char_set = string.ascii_letters + string.digits
-    length = random.randint(min_length, max_length)
-    return ''.join(map(lambda unused: random.choice(char_set), range(length)))
-
-
-def signal_handler(sig, frame):
-    global shutdown
-    if shutdown:  # pressed twice, so exit immediately
-        sys.exit(0)
-    shutdown = True  # pressed once, exit gracefully
-
-
-def unsupported_feature(feature=""):
-    print("Error: You have an unsupported version of Python interpreter dnspython library.")
-    print("       Some features such as DoT and DoH are not available. You should upgrade")
-    print("       the Python interpreter to at least 3.10 and reinstall dependencies.")
-    if feature:
-        print("Missing Feature: %s" % feature)
-    sys.exit(127)
-
-
-def valid_rdatatype(rtype):
+def valid_rdatatype(rtype: str) -> bool:
     # validate RR type
     try:
         _ = dns.rdatatype.from_text(rtype)
@@ -237,49 +229,35 @@ def valid_rdatatype(rtype):
     return True
 
 
-def flags_to_text(flags):
-    # Standard DNS flags
+# Standard DNS flags
+_DNS_FLAG_QR = 0x8000
+_DNS_FLAG_AA = 0x0400
+_DNS_FLAG_TC = 0x0200
+_DNS_FLAG_RD = 0x0100
+_DNS_FLAG_RA = 0x0080
+_DNS_FLAG_AD = 0x0020
+_DNS_FLAG_CD = 0x0010
 
-    QR = 0x8000
-    AA = 0x0400
-    TC = 0x0200
-    RD = 0x0100
-    RA = 0x0080
-    AD = 0x0020
-    CD = 0x0010
+_DNS_FLAGS_BY_TEXT = {
+    'QR': _DNS_FLAG_QR,
+    'AA': _DNS_FLAG_AA,
+    'TC': _DNS_FLAG_TC,
+    'RD': _DNS_FLAG_RD,
+    'RA': _DNS_FLAG_RA,
+    'AD': _DNS_FLAG_AD,
+    'CD': _DNS_FLAG_CD
+}
 
-    # EDNS flags
-    # DO = 0x8000
 
-    _by_text = {
-        'QR': QR,
-        'AA': AA,
-        'TC': TC,
-        'RD': RD,
-        'RA': RA,
-        'AD': AD,
-        'CD': CD
-    }
-
-    _by_value = dict([(y, x) for x, y in _by_text.items()])
-    # _flags_order = sorted(_by_value.items(), reverse=True)
-
-    _by_value = dict([(y, x) for x, y in _by_text.items()])
+def flags_to_text(flags: int) -> str:
+    _by_value = {value: key for key, value in _DNS_FLAGS_BY_TEXT.items()}
 
     order = sorted(_by_value.items(), reverse=True)
     text_flags = []
     for k, v in order:
-        if flags & k != 0:
+        if (flags & k) != 0:
             text_flags.append(v)
         else:
             text_flags.append('--')
 
     return ' '.join(text_flags)
-
-
-def setup_signal_handler():
-    try:
-        signal.signal(signal.SIGTSTP, signal.SIG_IGN)  # ignore CTRL+Z
-        signal.signal(signal.SIGINT, signal_handler)  # custom CTRL+C handler
-    except AttributeError:  # not all signals are supported on all platforms
-        pass
