@@ -25,6 +25,7 @@
 # OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 
+import concurrent.futures
 import datetime
 import getopt
 import ipaddress
@@ -33,8 +34,9 @@ import os
 import signal
 import socket
 import sys
+import threading
 import time
-from typing import List, Any
+from typing import List, Any, Optional
 
 import dns.rcode
 import dns.rdatatype
@@ -48,6 +50,7 @@ __author__ = 'Babak Farrokhi (babak@farrokhi.net)'
 __license__ = 'BSD'
 __progname__ = os.path.basename(sys.argv[0])
 shutdown = False
+print_lock = threading.Lock()
 
 
 def setup_signal_handler() -> None:
@@ -97,6 +100,102 @@ def maxlen(names: List[str]) -> int:
     return max(len(name) for name in names) if names else 0
 
 
+def evaluate_server(server: str, qname: str, rdatatype: str, waittime: int, count: int, proto: int,
+                    dst_port: int, src_ip: Optional[str], use_edns: bool, force_miss: bool, want_dnssec: bool,
+                    width: int, color: Colors, verbose: bool, json_output: bool, json_filename: str) -> str:
+    if not server.strip():
+        return ""
+
+    server = server.replace(' ', '')
+    resolver = ""
+    try:
+        ipaddress.ip_address(server)
+        resolver = server
+    except ValueError:
+        try:
+            results = socket.getaddrinfo(server, None, socket.AF_UNSPEC, socket.SOCK_DGRAM)
+            if results and len(results[0]) > 4 and results[0][4]:
+                resolver = str(results[0][4][0])
+            else:
+                return 'ERROR: cannot resolve hostname: %s' % server
+        except OSError:
+            return 'ERROR: cannot resolve hostname: %s' % server
+        except (IndexError, TypeError):
+            return 'ERROR: invalid address format for hostname: %s' % server
+
+    if not resolver:
+        return ""
+
+    try:
+        retval = dnsdiag.dns.ping(qname, resolver, dst_port, rdatatype, waittime, count, proto, src_ip,
+                                  use_edns=use_edns, force_miss=force_miss, want_dnssec=want_dnssec)
+
+    except (KeyboardInterrupt, SystemExit):
+        raise
+    except Exception as e:
+        return '%s: %s' % (server, e)
+
+    resolver = server.ljust(width + 1)
+    text_flags = flags_to_text(retval.flags)
+    edns_flags_text = dns.flags.edns_to_text(retval.ednsflags)
+    if edns_flags_text:
+        text_flags = " ".join([text_flags, edns_flags_text])
+    else:
+        text_flags = " ".join([text_flags, "--"])
+
+    s_ttl = str(retval.ttl) if retval.ttl is not None else "N/A"
+
+    if retval.r_lost_percent > 0:
+        l_color = color.O
+    else:
+        l_color = color.N
+
+    output_lines = []
+
+    if json_output:
+        outer_data = {
+            'hostname': qname,
+            'data': {
+                'hostname': qname,
+                'timestamp': str(datetime.datetime.now()),
+                'resolver': resolver.rstrip(),
+                'r_min': retval.r_min,
+                'r_avg': retval.r_avg,
+                'r_max': retval.r_max,
+                'r_stddev': retval.r_stddev,
+                'r_lost_percent': retval.r_lost_percent,
+                's_ttl': s_ttl,
+                'text_flags': text_flags,
+                'flags': retval.flags,
+                'ednsflags': retval.ednsflags,
+                'rcode': retval.rcode,
+                'rcode_text': retval.rcode_text,
+            }
+        }
+
+        if json_filename == '-':
+            output_lines.append(json.dumps(outer_data, indent=2))
+        else:
+            with print_lock:
+                with open(json_filename, 'a+') as outfile:
+                    json.dump(outer_data, outfile, indent=2)
+
+    else:
+        result = "%s  %-7.2f  %-7.2f  %-7.2f  %-10.2f  %s%%%-3d%s     %-7s  %-26s  %-12s" % (
+            resolver, retval.r_avg, retval.r_min, retval.r_max, retval.r_stddev, l_color, retval.r_lost_percent,
+            color.N, s_ttl, text_flags, retval.rcode_text)
+        output_lines.append(result.rstrip())
+
+    if verbose and retval.answer and not json_output:
+        ans_index = 1
+        for answer in retval.answer:
+            output_lines.append("Answer %d [ %s%s%s ]" % (ans_index, color.G, answer, color.N))
+            ans_index += 1
+        output_lines.append("")
+
+    return "\n".join(output_lines) if output_lines else ""
+
+
 def main() -> None:
     global shutdown
     setup_signal_handler()
@@ -279,105 +378,29 @@ def main() -> None:
                   '  avg(ms)  min(ms)  max(ms)  stddev(ms)  lost(%)  ttl      flags                      response')
             print((95 + width) * '-')
 
-        for server in f:
-            # Check for shutdown signal
-            if shutdown:
-                break
+        max_workers = min(len(f), 10)
+        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
+            future_to_server = {}
+            for server in f:
+                if shutdown:
+                    break
+                future = executor.submit(evaluate_server, server, qname, rdatatype, waittime, count, proto,
+                                       dst_port, src_ip, use_edns, force_miss, want_dnssec, width, color,
+                                       verbose, json_output, json_filename if json_output else '')
+                future_to_server[future] = server
 
-            # check if we have a valid dns server address
-            if not server.strip():
-                continue
-            server = server.replace(' ', '')
-            resolver = ""
-            try:
-                ipaddress.ip_address(server)
-                resolver = server
-            except ValueError:
+            for future in future_to_server:
+                if shutdown:
+                    break
                 try:
-                    results = socket.getaddrinfo(server, None, socket.AF_UNSPEC, socket.SOCK_DGRAM)
-                    if results and len(results[0]) > 4 and results[0][4]:
-                        resolver = str(results[0][4][0])
-                    else:
-                        print('ERROR: cannot resolve hostname:', server)
-                        continue
-                except OSError:
-                    print('ERROR: cannot resolve hostname:', server)
-                    continue
-                except (IndexError, TypeError):
-                    print('ERROR: invalid address format for hostname:', server)
-                    continue
-
-            if not resolver:
-                continue
-
-            try:
-                retval = dnsdiag.dns.ping(qname, resolver, dst_port, rdatatype, waittime, count, proto, src_ip,
-                                          use_edns=use_edns, force_miss=force_miss, want_dnssec=want_dnssec)
-
-            except KeyboardInterrupt:
-                shutdown = True
-                break
-            except SystemExit:
-                shutdown = True
-                break
-            except Exception as e:
-                print('%s: %s' % (server, e))
-                continue
-
-            resolver = server.ljust(width + 1)
-            text_flags = flags_to_text(retval.flags)
-            edns_flags_text = dns.flags.edns_to_text(retval.ednsflags)
-            if edns_flags_text:
-                text_flags = " ".join([text_flags, edns_flags_text])
-            else:
-                text_flags = " ".join([text_flags, "--"])
-
-            s_ttl = str(retval.ttl) if retval.ttl is not None else "N/A"
-
-            if retval.r_lost_percent > 0:
-                l_color = color.O
-            else:
-                l_color = color.N
-
-            if json_output:
-                outer_data = {
-                    'hostname': qname,
-                    'data': {
-                        'hostname': qname,
-                        'timestamp': str(datetime.datetime.now()),
-                        'resolver': resolver.rstrip(),
-                        'r_min': retval.r_min,
-                        'r_avg': retval.r_avg,
-                        'r_max': retval.r_max,
-                        'r_stddev': retval.r_stddev,
-                        'r_lost_percent': retval.r_lost_percent,
-                        's_ttl': s_ttl,
-                        'text_flags': text_flags,
-                        'flags': retval.flags,
-                        'ednsflags': retval.ednsflags,
-                        'rcode': retval.rcode,
-                        'rcode_text': retval.rcode_text,
-                    }
-                }
-
-                if json_filename == '-':  # stdout
-                    print(json.dumps(outer_data, indent=2))
-                else:
-                    with open(json_filename, 'a+') as outfile:
-                        json.dump(outer_data, outfile, indent=2)
-
-            else:
-                result = "%s  %-7.2f  %-7.2f  %-7.2f  %-10.2f  %s%%%-3d%s     %-7s  %-26s  %-12s" % (
-                    resolver, retval.r_avg, retval.r_min, retval.r_max, retval.r_stddev, l_color, retval.r_lost_percent,
-                    color.N, s_ttl, text_flags, retval.rcode_text)
-                print(result.rstrip(), flush=True)
-
-            if verbose and retval.answer and not json_output:
-                ans_index = 1
-                for answer in retval.answer:
-                    print("Answer %d [ %s%s%s ]" % (ans_index, color.G, answer, color.N))
-                    ans_index += 1
-                print("")
+                    result_output = future.result()
+                    if result_output:
+                        print(result_output, flush=True)
+                except (KeyboardInterrupt, SystemExit):
+                    shutdown = True
+                    break
+                except Exception:
+                    pass
 
     except Exception as e:
         die(f'{server}: {e}')
